# -*- coding: utf-8 -*-
# -*- mode: python -*-
"""spike detection and extraction

Copyright (C) 2013 Dan Meliza <dmeliza@gmail.com>
Created Fri Jul 12 14:05:16 2013
"""
from typing import Iterable, Union, Tuple, Optional, Iterator, Mapping
import numpy as np

numeric = Union[int, float, np.number]


def filter_times(
    times: Iterable[numeric], min: numeric, max: numeric
) -> Tuple[numeric]:
    return tuple(t for t in times if (t > min) and (t < max))


def realign_spikes(
    times: Iterable[numeric],
    spikes: np.ndarray,
    upsample: int,
    jitter: int = 3,
    reflect_fft: bool = False,
    expected_peak: Optional[int] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """Realign spikes to their peaks using bandwidth-limited resampling

    times    : one-dimensional array of spike times, in units of samples
    spikes   : array of spike waveforms, with dimensions (nspikes, npoints)
    upsample : integer, the factor by which to upsample the data for realignment (must be greater than 1)
    jitter   : samples (in original data) to search for true peak
    expected_peak : if supplied, searches around this index for the peak. If not supplied,
                    uses the argmax of the mean spike

    Returns (times, spikes), with the sampling rate increased by a factor of upsample

    """
    assert (
        isinstance(upsample, int) and upsample > 1
    ), "Upsampling factor must be an integer greater than 1"
    nevents, nsamples = spikes.shape

    # first infer the expected peak time
    if expected_peak is None:
        expected_peak = np.mean(spikes, 0).argmax() * upsample
    else:
        expected_peak *= upsample
    spikes = fftresample(spikes, int(nsamples * upsample), reflect=reflect_fft)
    # find peaks within upsample samples of mean peak
    shift = find_peaks(spikes, expected_peak, upsample * jitter)
    start = shift + upsample * jitter
    nshifted = (nsamples - 2 * jitter) * upsample
    shifted = np.zeros((nevents, nshifted))
    for i, spike in enumerate(spikes):
        shifted[i, :] = spike[start[i] : start[i] + nshifted]
    return (np.asarray(times) * upsample + shift, shifted)


def find_peaks(spikes: np.ndarray, peak: int, window: int) -> np.ndarray:
    """Locate the peaks in an array of spikes.

    spikes: resampled spike waveforms, dimensions (nspikes, nsamples)
    peak:   the expected peak location
    window: the number of samples to either side of the peak to look for the peak

    Returns array of shift values relative to peak

    """
    r = slice(peak - window, peak + window + 1)
    return spikes[:, r].argmax(1) - window


def fftresample(S: np.ndarray, npoints: int, reflect: bool = False) -> np.ndarray:
    """Resample an array of waveforms using discrete fourier transform. The
    signal is transformed in the fourier domain and then padded or truncated to
    the desired size. This should be equivalent to a sinc resampling.

    S: a 2D array of waveforms (nspikes, nsamples)
    npoints: the desired duration of the resampled signal
    reflect: if True, pad the sample with reflected copies on either end to suppress edge effects

    """
    from numpy.fft import rfft, irfft

    if reflect:
        Srev = S[:, ::-1]
        S = np.column_stack([Srev, S, Srev])
        npoints = npoints * 3
    Sf = rfft(S, axis=1)
    Srs = (1.0 * npoints / S.shape[1]) * irfft(Sf, npoints, axis=1)
    if reflect:
        npoints = npoints // 3
        return Srs[:, npoints : npoints * 2]
    else:
        return Srs


def trim_waveforms(
    spikes: Iterable[np.ndarray], times: Iterable[int], peak_t: int, n_rise: int
) -> Iterator[Tuple[int, np.ndarray]]:
    """Trims spike waveforms to remove overlapping peaks.

    The peaks() function is used to extract waveforms in a window surrounding
    peak times. The windows for successive peaks may overlap each other if the
    peak times are closer together than the window size, which can result in
    secondary peaks occuring later in the window corresponding to the subsequent
    peak(s). If this is not desired, this function can be used to trim the
    samples after n_rise before the next peak.

    times: the times of the spikes (in samples)
    spikes: a list of spike waveforms, or a 2D array with dimensions (nspikes, nsamples)
    peak_t: the index of the peak in the spike waveforms
    n_rise: trim up to this many samples before the subsequent peak

    Note that there is nothing to prevent a spike from being completely trimmed
    if the subsequent spike is very close and n_rise is too large. In practice
    this is highly unlikely.

    """
    diffs = np.diff(times)
    for t, dt, s in zip(times, diffs, spikes):
        i_last = min(
            peak_t + dt - n_rise,
            # s[peak_t:].argmin() + peak_t,
            s.size,
        )
        yield t, s[:i_last]
    # the last spike
    s = spikes[-1]
    # i_last = min(s[peak_t:].argmin() + peak_t, s.size)
    yield times[-1], s[: s.size]


def runlength_encode(arr: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Run length encoding of arr. Returns (runlengths, startpositions, values)"""
    # from stackoverflow 1066758
    n = len(arr)
    if n == 0:
        return (None, None, None)
    else:
        y = np.array(arr[1:] != arr[:-1])  # pairwise unequal (string safe)
        i = np.append(np.where(y), n - 1)  # must include last element posi
        z = np.diff(np.append(-1, i))  # run lengths
        p = np.cumsum(np.append(0, z))[:-1]  # positions
        return (z, p, arr[i])



# -*- coding: utf-8 -*-
# -*- mode: python -*-
import unittest

import numpy as np

from quickspikes.spikes import detector, peaks, find_run
from quickspikes.tools import filter_times, realign_spikes, trim_waveforms
from quickspikes.intracellular import SpikeFinder, spike_shape

# a nice surrogate spike with 20 samples before peak and 40 after
a_spike = np.array([-1290,  -483,  -136,  -148,  -186,   637,   328,    41,    63,
                    42,   377,   872,   639,   -17,   538,   631,   530,   693,
                    743,  3456,  6345,  5868,  4543,  3087,  1691,   830,   241,
                    -350,  -567,  -996,  -877, -1771, -1659, -1968, -2013, -2290,
                    -2143, -1715, -1526, -1108,  -500,   333,    25,  -388,  -368,
                    -435,  -817,  -858,  -793, -1089,   -16,  -430,  -529,  -252,
                    -3,  -786,   -47,  -266,  -963,  -365], dtype=np.int16)
t_peak = a_spike.argmax()
t_trough = a_spike.argmin()
a_recording_dbl = np.zeros(10000, dtype='d')
a_recording_int = np.zeros(10000, dtype=a_spike.dtype)
a_times = [100, 400, 1200, 1500, 5000, 5200, 6123, 9730]
for t in a_times:
    a_recording_dbl[t:t + a_spike.size] += a_spike
    a_recording_int[t:t + a_spike.size] += a_spike

b_recording = np.load("test/intra_spike.npy")
b_times = [7635, 8412, 9363, 10424, 11447, 12661, 13887, 15079, 16373,
           17753, 19168, 20682, 22357, 23979, 25574, 27209, 28989,
           30508, 32088, 33778]
b_takeoff = 24

c_recording = np.load("test/intra_spike_narrow.npy")
c_times = [8325, 8816, 9368, 9985, 10619, 11276, 11968, 12610, 13240, 13900, 14485, 15193, 15840, 16601]
c_takeoff = 14


class TestQuickspikes(unittest.TestCase):

    def test_detect_extrac_spikes(self):
        det = detector(2000, 40)
        self.assertSequenceEqual(det.send(a_recording_dbl), [t + t_peak for t in a_times])
        self.assertSequenceEqual(det.send(-a_recording_dbl), [t + t_trough for t in a_times])
        self.assertSequenceEqual(det(a_recording_dbl), [t + t_peak for t in a_times])

    def test_extract_spikes_nofilter(self):
        with self.assertRaises(ValueError):
            peaks(a_recording_dbl, [t + t_peak for t in a_times], n_before=20, n_after=300)

    def test_extract_spikes_double(self):
        n_before = 20
        n_after = 300
        times = filter_times([t + t_peak for t in a_times], n_before, a_recording_dbl.size - n_after)
        x = peaks(a_recording_dbl, times, n_before=n_before, n_after=n_after)
        # last peak should get dropped
        self.assertEqual(x.shape[0], len(a_times) - 1)
        self.assertEqual(x.shape[1], 320)
        self.assertTrue(np.all(a_spike == x[0,:a_spike.size]))

    def test_extract_spikes_integer(self):
        n_before = 20
        n_after = 300
        times = filter_times([t + t_peak for t in a_times], n_before, a_recording_int.size - n_after)
        x = peaks(a_recording_int, times, n_before=n_before, n_after=n_after)
        # last peak should get dropped
        self.assertEqual(x.shape[0], len(a_times) - 1)
        self.assertEqual(x.shape[1], 320)
        self.assertTrue(np.all(a_spike == x[0,:a_spike.size]))

    def test_detect_intrac_spikes(self):
        det = detector(0, 100)
        self.assertSequenceEqual(det.send(b_recording), b_times)
        det = detector(-20, 100)
        self.assertSequenceEqual(det.send(b_recording), b_times)

    def test_align_spikes(self):
        jitter = 4
        spikes = peaks(b_recording, b_times, 200, 400)
        times, aligned = realign_spikes(b_times, spikes, 3, jitter)
        apeak = aligned.argmax(1)
        self.assertTrue((apeak == apeak[0]).all())
        # times should have shifted no more than jitter
        self.assertEqual(len(b_times), times.size)
        for t1, t2 in zip(b_times, times):
            self.assertAlmostEqual(t1, t2 // 3, delta=jitter)

    def test_detect_intrac_spikes_narrow(self):
        det = detector(-20, 100)
        self.assertSequenceEqual(det.send(c_recording), c_times)


class TestTools(unittest.TestCase):

    def test_trim(self):
        # empirical test: windows should have no more than one peak after
        # trimming
        det = detector(-20, 100)
        spikes = peaks(c_recording, c_times, 200, 700)
        for _, spike_w in trim_waveforms(spikes, c_times, 200, 100):
            t = det(spike_w.astype("d"))
            self.assertEqual(len(t), 1)
            self.assertSequenceEqual(t, [200])

    def test_run(self):
        a = np.arange(-10, 10)
        self.assertEqual(find_run(a, 0, 5), 11)

    def test_no_run(self):
        a = np.zeros(20)
        a[10:14] = 1
        self.assertEqual(find_run(a, 0, 5), None)


class TestSpikeShape(unittest.TestCase):

    def test_extrac_shape(self):
        shape = spike_shape(a_spike, 1)
        self.assertEqual(shape.peak_t, t_peak)
        self.assertEqual(t_peak + shape.trough_t, t_trough)

    def test_intrac_trough(self):
        spikes = peaks(b_recording, b_times, 200, 400)
        for i, spike in enumerate(spikes):
            shape = spike_shape(spike, dt=1, t_baseline=100, min_rise=13)
            peak = shape.peak_t
            self.assertEqual(shape.trough_t, spike[peak:].argmin())

    def test_intrac_narrow_trough(self):
        spikes = peaks(c_recording, c_times, 200, 400)
        for spike in spikes:
            shape = spike_shape(spike, dt=1, t_baseline=100, min_rise=13)
            peak = shape.peak_t
            self.assertEqual(shape.trough_t, spike[peak:].argmin())

    def test_intrac_trough_no_min(self):
        spikes = peaks(b_recording, b_times[:1], 100, 100)
        for spike in spikes:
            shape = spike_shape(spike, dt=1, t_baseline=100, min_rise=13)
            peak = shape.peak_t
            self.assertEqual(shape.trough_t, spike[peak:].argmin())

    def test_intrac_onset(self):
        # this case is based on manual inspection of the spike waveform
        spike = peaks(b_recording, b_times[:1], 200, 100)[0]
        shape = spike_shape(spike, dt=1, t_baseline=100, min_rise=13)
        self.assertEqual(shape.takeoff_t, b_takeoff)

    def test_intrac_no_onset(self):
        spike = peaks(b_recording, b_times[:1], 100, 100)[0]
        shape = spike_shape(spike, dt=1, t_baseline=80, min_rise=13)
        self.assertIsNone(shape.takeoff_t)

    def test_intrac_narrow_onset(self):
        # this case is based on manual inspection of the spike waveform
        spike = peaks(c_recording, c_times[:1], 200, 100)[0]
        shape = spike_shape(spike, dt=1, t_baseline=100, min_rise=13)
        self.assertEqual(shape.takeoff_t, c_takeoff)

    def test_bad_spike_peak_at_edge(self):
        """ spike_shape returns None if max is at the edge of the waveform """
        bad = np.arange(100, -100, -1)
        self.assertIsNone(spike_shape(bad, dt=1, t_baseline=100, min_rise=13))
        self.assertIsNone(spike_shape(bad[::-1], dt=1, t_baseline=100, min_rise=13))


class TestDynamicExtractor(unittest.TestCase):

    def test_intrac_broad(self):
        detector = SpikeFinder(50, 350, 5000)
        shape = detector.calculate_threshold(b_recording)
        self.assertEqual(shape.takeoff_t, b_takeoff)
        for i, (time, spike) in enumerate(
                detector.extract_spikes(b_recording, 10, upsample=2, jitter=4)
        ):
            self.assertAlmostEqual(time, b_times[i], delta=4)

    def test_intrac_narrow(self):
        detector = SpikeFinder(50, 350, 5000)
        shape = detector.calculate_threshold(c_recording)
        self.assertEqual(shape.takeoff_t, c_takeoff)
        for i, (time, spike) in enumerate(
            detector.extract_spikes(c_recording, 10, upsample=2, jitter=4)
        ):
            self.assertAlmostEqual(time, c_times[i], delta=4)

    def test_no_spikes(self):
        """ extract_spikes should yield nothing if there are no spikes detected """
        # In some rare cases, there will be a single spike that passes the first spike
        # threshold but is not detected due to the minimum height requirement
        # (thresh_min).
        detector = SpikeFinder(50, 350, 5000)
        signal = np.random.randn(100000)
        detector.spike_thresh = 50
        times = [t for t, s in detector.extract_spikes(signal, 40)]
        self.assertSequenceEqual(times, [])

    def test_thresh_for_max_at_edge_of_signal(self):
        """ calculate_threshold should ignore spikes that are too close to the edge of the signal """
        detector = SpikeFinder(50, 350, 5000)
        clip = b_times[0] - 300
        _ = detector.calculate_threshold(b_recording[clip:])
        peaks = detector.detect_spikes(b_recording[clip:])
        self.assertEqual(peaks[0] + clip, b_times[1])
